# -*- coding: utf-8 -*-
import copy
import time
import hashlib

from amplify.agent.containers.nginx.config.parser import NginxConfigParser
from amplify.agent.util import subp
from amplify.agent.context import context

__author__ = "Mike Belov"
__copyright__ = "Copyright (C) 2015, Nginx Inc. All rights reserved."
__credits__ = ["Mike Belov", "Andrei Belov", "Ivan Poluyanov", "Oleg Mamontov", "Andrew Alexeev"]
__license__ = ""
__maintainer__ = "Mike Belov"
__email__ = "dedm@nginx.com"


class NginxConfig(object):
    """
    Nginx config representation
    Parses configs with all includes, etc

    Main tasks:
    - find all log formats
    - find all access logs
    - find all error logs
    - find stub_status url
    """

    def __init__(self, filename, binary=None, prefix=None):
        self.filename = filename
        self.binary = binary
        self.prefix = prefix
        self.log_formats = {}
        self.access_logs = {}
        self.error_logs = []
        self.stub_status = []
        self.plus_status = []
        self.parser = NginxConfigParser(filename)

        # initial parsing
        self.tree = self.parser.tree
        self.files = self.parser.files
        self.index = self.parser.index
        self.parser_errors = self.parser.errors

        # go through and collect all logical data
        self.__recursive_search(subtree=self.parser.simplify())

    def __recursive_search(self, subtree=None, ctx=None):
        """
        Searches needed data in config's tree

        :param subtree: dict with tree to parse
        :param ctx: dict with context
        """
        ctx = ctx if ctx is not None else {}
        subtree = subtree if subtree is not None else {}

        for key, value in subtree.iteritems():
            if key == 'error_log':
                log_name = value.split(' ')[0]
                if not log_name.startswith('/'):
                    log_name = '%s/%s' % (self.prefix, log_name)
                self.error_logs.append(log_name)
            elif key == 'access_log':
                parts = value.split(' ')
                log_format = None if len(parts) == 1 else parts[-1]
                log_name = parts[0]
                if not log_name.startswith('/'):
                    log_name = '%s/%s' % (self.prefix, log_name)
                self.access_logs[log_name] = log_format
            elif key == 'log_format':
                for k, v in value.iteritems():
                    self.log_formats[k] = v
            elif key == 'server' and isinstance(value, list) and 'upstream' not in ctx:
                for server in value:
                    current_ctx = copy.copy(ctx)
                    listen = server.get('listen', [])
                    listen = listen if isinstance(listen, list) else [listen]
                    if len(listen):
                        listen_first_part = listen[0].split(' ')[0]
                        if listen_first_part.startswith('['):
                            # ipv6
                            addr, port = listen_first_part.rsplit(':', 1)
                            ctx['port'] = port
                            ctx['server'] = server.get('server_name')
                        elif ':' in listen_first_part:
                            addr, port = listen_first_part.split(':')
                            ctx['port'] = port
                            ctx['server'] = addr
                        else:
                            ctx['port'] = listen_first_part
                            ctx['server'] = server.get('server_name')
                    self.__recursive_search(subtree=server, ctx=ctx)
                    ctx = current_ctx
            elif key == 'upstream':
                for upstream, upstream_info in value.iteritems():
                    current_ctx = copy.copy(ctx)
                    ctx['upstream'] = upstream
                    self.__recursive_search(subtree=upstream_info, ctx=ctx)
                    ctx = current_ctx
            elif key == 'location':
                for location, location_info in value.iteritems():
                    current_ctx = copy.copy(ctx)
                    ctx['location'] = location
                    self.__recursive_search(subtree=location_info, ctx=ctx)
                    ctx = current_ctx
            elif key == 'stub_status' and value == 'on' and ctx and 'port' in ctx and 'server' in ctx:
                self.stub_status.append(self.__status_url(ctx))
            elif key == 'status' and ctx and 'port' in ctx and 'server' in ctx:
                self.plus_status.append(self.__status_url(ctx))
            elif isinstance(value, dict):
                self.__recursive_search(subtree=value, ctx=ctx)
            elif isinstance(value, list):
                for next_subtree in value:
                    if isinstance(next_subtree, dict):
                        self.__recursive_search(subtree=next_subtree, ctx=ctx)

    def __status_url(self, ctx):
        location = ctx.get('location', '/')

        # remove all modifiers
        location_parts = location.split(' ')
        final_location_part = location_parts[-1]

        # url build
        server = ctx.get('server') or 'localhost'
        port = ctx.get('port') or '80'
        return '%s:%s%s' % (server, port, final_location_part)

    def run_test(self):
        """
        Tests the configuration using nginx -t
        Saves event info if syntax check was not successful
        """
        result = []
        start_time = time.time()
        if self.binary:
            try:
                _, nginx_t_err = subp.call("%s -t" % self.binary, check=False)
                for line in nginx_t_err:
                    if 'syntax is' in line and 'syntax is ok' not in line:
                        result.append(line)
            except Exception, e:
                exception_name = e.__class__.__name__
                context.log.error('failed to %s -t due to %s' % (self.binary, exception_name))
                context.log.debug('additional info:', exc_info=True)
        end_time = time.time()
        return result, end_time - start_time

    def checksum(self):
        """
        Calculates total checksum of all config/files files

        :param config: NginxConfig object
        :return: str checksum
        """
        checksums = []
        for filename, filedata in self.files.iteritems():
            checksums.append(hashlib.sha256(open(filename).read()).hexdigest())
        return hashlib.sha256('.'.join(checksums)).hexdigest()
